@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003-2016 GraphicsMagick Group
+% Copyright (C) 2003-2018 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 % Copyright 1991-1999 E. I. du Pont de Nemours and Company
 %
@@ -1286,7 +1286,7 @@ static Image *ReadMIFFImage(const ImageInfo *image_info,
 
     (void) LogMagickEvent(CoderEvent,GetMagickModule(),
                           "id=\"%s\" version=%g class=%s colorspace=%s compression=%s matte=%s "
-			  "columns=%lu rows=%lu depth=%u",
+                          "columns=%lu rows=%lu depth=%u",
                           id,version,
                           ClassTypeToString(image->storage_class),
                           ColorspaceTypeToString(image->colorspace),
@@ -1355,10 +1355,12 @@ static Image *ReadMIFFImage(const ImageInfo *image_info,
               p=image->directory+strlen(image->directory);
             }
           c=ReadBlobByte(image);
+          if (c == EOF)
+            break;
           *p++=c;
         } while (c != '\0');
       }
- 
+
     /*
       Attached profiles.
     */
@@ -1531,8 +1533,8 @@ static Image *ReadMIFFImage(const ImageInfo *image_info,
 #if defined(HasZLIB)
       case ZipCompression:
         {
-	  int
-	    code=0;
+          int
+            code=0;
 
           for (y=0; y < (long) image->rows; y++)
             {
@@ -1608,7 +1610,7 @@ static Image *ReadMIFFImage(const ImageInfo *image_info,
                 if (QuantumTick(y,image->rows))
                   if (!MagickMonitorFormatted(y,image->rows,exception,
                                               LoadImageText,image->filename,
-					      image->columns,image->rows))
+                                              image->columns,image->rows))
                     break;
             }
           break;
@@ -1617,8 +1619,8 @@ static Image *ReadMIFFImage(const ImageInfo *image_info,
 #if defined(HasBZLIB)
       case BZipCompression:
         {
-	  int
-	    code=0;
+          int
+            code=0;
 
           for (y=0; y < (long) image->rows; y++)
             {
@@ -1685,7 +1687,7 @@ static Image *ReadMIFFImage(const ImageInfo *image_info,
                 if (QuantumTick(y,image->rows))
                   if (!MagickMonitorFormatted(y,image->rows,exception,
                                               LoadImageText,image->filename,
-					      image->columns,image->rows))
+                                              image->columns,image->rows))
                     break;
             }
           break;
@@ -1715,7 +1717,7 @@ static Image *ReadMIFFImage(const ImageInfo *image_info,
                 if (QuantumTick(y,image->rows))
                   if (!MagickMonitorFormatted(y,image->rows,exception,
                                               LoadImageText,image->filename,
-					      image->columns,image->rows))
+                                              image->columns,image->rows))
                     break;
 
             }
@@ -1737,7 +1739,7 @@ static Image *ReadMIFFImage(const ImageInfo *image_info,
                 if (QuantumTick(y,image->rows))
                   if (!MagickMonitorFormatted(y,image->rows,exception,
                                               LoadImageText,image->filename,
-					      image->columns,image->rows))
+                                              image->columns,image->rows))
                     break;
             }
           break;
@@ -2150,13 +2152,13 @@ static unsigned int WriteMIFFImage(const ImageInfo *image_info,Image *image)
 
   ImageProfileIterator
     profile_iterator;
-  
+
   const char
     *profile_name;
-  
+
   const unsigned char
     *profile_info;
-  
+
   size_t
     profile_length;
 
@@ -2165,6 +2167,9 @@ static unsigned int WriteMIFFImage(const ImageInfo *image_info,Image *image)
     zip_info;
 #endif
 
+  size_t
+    image_list_length;
+
   /*
     Open output image file.
   */
@@ -2172,6 +2177,7 @@ static unsigned int WriteMIFFImage(const ImageInfo *image_info,Image *image)
   assert(image_info->signature == MagickSignature);
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
+  image_list_length=GetImageListLength(image);
   status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
   if (status == False)
     ThrowWriterException(FileOpenError,UnableToOpenFile,image);
@@ -2589,8 +2595,8 @@ static unsigned int WriteMIFFImage(const ImageInfo *image_info,Image *image)
 #if defined(HasZLIB)
         case ZipCompression:
         {
-	  int
-	    code;
+          int
+            code;
 
           if (y == 0)
             {
@@ -2638,8 +2644,8 @@ static unsigned int WriteMIFFImage(const ImageInfo *image_info,Image *image)
 #if defined(HasBZLIB)
         case BZipCompression:
         {
-	  int
-	    code;
+          int
+            code;
 
           if (y == 0)
             {
@@ -2726,7 +2732,7 @@ static unsigned int WriteMIFFImage(const ImageInfo *image_info,Image *image)
         if (QuantumTick(y,image->rows))
           if (!MagickMonitorFormatted(y,image->rows,&image->exception,
                                       SaveImageText,image->filename,
-				      image->columns,image->rows))
+                                      image->columns,image->rows))
             break;
     }
     MagickFreeMemory(pixels);
@@ -2734,7 +2740,7 @@ static unsigned int WriteMIFFImage(const ImageInfo *image_info,Image *image)
     if (image->next == (Image *) NULL)
       break;
     image=SyncNextImageInList(image);
-    status=MagickMonitorFormatted(scene++,GetImageListLength(image),
+    status=MagickMonitorFormatted(scene++,image_list_length,
                                   &image->exception,SaveImagesText,
                                   image->filename);
     if (status == False)
