@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003-2013 GraphicsMagick Group
+% Copyright (C) 2003-2018 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 %
 % This program is covered by multiple licenses, which are described in
@@ -125,13 +125,28 @@ static MagickBool IsMPC(const unsigned char *magick,const size_t length)
 %
 %
 */
+
+#define ThrowMPCReaderException(code_,reason_,image_) \
+do { \
+  if (number_of_profiles > 0) \
+    { \
+      unsigned int _index; \
+      for (_index=0; _index < number_of_profiles; _index++) \
+        { \
+          MagickFreeMemory(profiles[_index].name); \
+          MagickFreeMemory(profiles[_index].info); \
+        } \
+      MagickFreeMemory(profiles); \
+      number_of_profiles=0; \
+    } \
+  ThrowReaderException(code_,reason_,image_); \
+} while (0);
 static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
 {
   char
     cache_filename[MaxTextExtent],
     id[MaxTextExtent],
-    keyword[MaxTextExtent],
-    *values;
+    keyword[MaxTextExtent];
 
   ExtendedSignedIntegralType
     offset;
@@ -148,9 +163,6 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
   register unsigned char
     *p;
 
-  size_t
-    length;
-
   unsigned int
     status;
 
@@ -199,11 +211,10 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
     /*
       Decode image header;  header terminates one character beyond a ':'.
     */
-    length=MaxTextExtent;
-    values=AllocateString((char *) NULL);
     quantum_depth=QuantumDepth;
     image->depth=8;
     image->compression=NoCompression;
+    image->storage_class=DirectClass;
     while (isgraph(c) && (c != ':'))
     {
       register char
@@ -214,22 +225,28 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
           char
             *comment;
 
+          size_t
+            comment_length;
+
           /*
             Read comment-- any text between { }.
           */
-          length=MaxTextExtent;
-          comment=AllocateString((char *) NULL);
+          comment_length=MaxTextExtent;
+          comment=MagickAllocateMemory(char *,comment_length);
+          if (comment == (char *) NULL)
+            ThrowMPCReaderException(ResourceLimitError,MemoryAllocationFailed,
+              image);
           p=comment;
           for ( ; comment != (char *) NULL; p++)
           {
             c=ReadBlobByte(image);
             if ((c == EOF) || (c == '}'))
               break;
-            if ((unsigned long) (p-comment+1) >= length)
+            if ((size_t) (p-comment+1) >= comment_length)
               {
                 *p='\0';
-                length<<=1;
-                MagickReallocMemory(char *,comment,length);
+                comment_length<<=1;
+                MagickReallocMemory(char *,comment,comment_length);
                 if (comment == (char *) NULL)
                   break;
                 p=comment+strlen(comment);
@@ -237,7 +254,7 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
             *p=c;
           }
           if (comment == (char *) NULL)
-            ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,
+            ThrowMPCReaderException(ResourceLimitError,MemoryAllocationFailed,
               image);
           *p='\0';
           (void) SetImageAttribute(image,"comment",comment);
@@ -247,40 +264,71 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
       else
         if (isalnum(c))
           {
+            char
+              *values;
+
+            size_t
+              values_length;
+
+            MagickBool
+              in_brace=MagickFalse;
+
             /*
-              Determine a keyword and its value.
+              Get keyword.
             */
+            keyword[0]='\0';
             p=keyword;
             do
             {
               if ((p-keyword) < (MaxTextExtent-1))
                 *p++=c;
               c=ReadBlobByte(image);
-            } while (isalnum(c) || (c == '-'));
+            } while ((c != '=') && (c != EOF));
             *p='\0';
-            while (isspace(c) || (c == '='))
+            if (c == EOF)
+              ThrowMPCReaderException(CorruptImageWarning,ImproperImageHeader,image);
+
+            /*
+              Get values.
+
+              Values not containing spaces are terminated by the first
+              white-space (or new-line) enountered.  Values containing
+              spaces and/or new-lines must be surrounded by braces.
+            */
+            values_length=MaxTextExtent;
+            values=MagickAllocateMemory(char *,values_length);
+            if (values == (char *) NULL)
+              ThrowMPCReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+            values[0]='\0';
+            c=ReadBlobByte(image);
+            in_brace=(c == '{');
+            if (in_brace)
               c=ReadBlobByte(image);
             p=values;
-            while ((c != '}') && (c != EOF))
-            {
-              if ((p-values+1) >= (int) length)
-                {
-                  *p='\0';
-                  length<<=1;
-                  MagickReallocMemory(char *,values,length);
-                  if (values == (char *) NULL)
+            while ((((!in_brace) && (c != '\n')) ||
+                     ((in_brace) && (c != '}'))) &&
+                   (c != EOF))
+              {
+                if ((size_t) (p-values+1) >= values_length)
+                  {
+                    *p='\0';
+                    values_length<<=1;
+                    MagickReallocMemory(char *,values,values_length);
+                    if (values == (char *) NULL)
+                      break;
+                    p=values+strlen(values);
+                  }
+                if (values == (char *) NULL)
+                  ThrowMPCReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+                *p++=c;
+                c=ReadBlobByte(image);
+                if (!in_brace)
+                  if (isspace(c))
                     break;
-                  p=values+strlen(values);
-                }
-              if (values == (char *) NULL)
-                ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
-              *p++=c;
-              c=ReadBlobByte(image);
-              if (*values != '{')
-                if (isspace(c))
-                  break;
-            }
+              }
             *p='\0';
+            (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                                  "keyword=\"%s\" values=\"%s\"",keyword,values);
             /*
               Assign a value to the specified keyword.
             */
@@ -428,8 +476,8 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
                   }
                 if (LocaleCompare(keyword,"grayscale") == 0)
                   {
-		    if (LocaleCompare(values,"True") == 0)
-		      image->is_grayscale=MagickTrue;
+                    if (LocaleCompare(values,"True") == 0)
+                      image->is_grayscale=MagickTrue;
                     break;
                   }
                 if (LocaleCompare(keyword,"green-primary") == 0)
@@ -487,8 +535,8 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
                   }
                 if (LocaleCompare(keyword,"monochrome") == 0)
                   {
-		    if (LocaleCompare(values,"True") == 0)
-		      image->is_monochrome=MagickTrue;
+                    if (LocaleCompare(values,"True") == 0)
+                      image->is_monochrome=MagickTrue;
                     break;
                   }
                 if (LocaleCompare(keyword,"montage") == 0)
@@ -537,7 +585,10 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
                     i=(long) number_of_profiles;
                     MagickReallocMemory(ProfileInfo *,profiles,(i+1)*sizeof(ProfileInfo));
                     if (profiles == (ProfileInfo *) NULL)
-                      ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+                      {
+                        MagickFreeMemory(values);
+                        ThrowMPCReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+                      }
                     profiles[i].name=AllocateString(keyword+8);
                     profiles[i].length=MagickAtoL(values);
                     profiles[i].info=(unsigned char *) NULL;
@@ -651,14 +702,25 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
                 break;
               }
             }
+            MagickFreeMemory(values);
           }
         else
-          c=ReadBlobByte(image);
+          {
+            c=ReadBlobByte(image);
+          }
       while (isspace(c))
         c=ReadBlobByte(image);
     }
-    MagickFreeMemory(values);
     (void) ReadBlobByte(image);
+
+    (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                          "id=\"%s\" class=%s compression=%s matte=%s "
+                          "columns=%lu rows=%lu depth=%u",
+                          id,ClassTypeToString(image->storage_class),
+                          CompressionTypeToString(image->compression),
+                          MagickBoolToString(image->matte),
+                          image->columns, image->rows, image->depth);
+
     /*
       Verify that required image information is defined.
     */
@@ -666,9 +728,9 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
         (image->storage_class == UndefinedClass) ||
         (image->compression == UndefinedCompression) || (image->columns == 0) ||
         (image->rows == 0))
-      ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
+      ThrowMPCReaderException(CorruptImageError,ImproperImageHeader,image);
     if (quantum_depth != QuantumDepth)
-      ThrowReaderException(CacheError,InconsistentPersistentCacheDepth,image);
+      ThrowMPCReaderException(CacheError,InconsistentPersistentCacheDepth,image);
     if (image->montage != (char *) NULL)
       {
         register char
@@ -679,7 +741,7 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
         */
         image->directory=AllocateString((char *) NULL);
         if (image->directory == (char *) NULL)
-          ThrowReaderException(CorruptImageError,UnableToReadImageData,image);
+          ThrowMPCReaderException(CorruptImageError,UnableToReadImageData,image);
         p=image->directory;
         do
         {
@@ -692,11 +754,13 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
               MagickReallocMemory(char *,image->directory,
                 (strlen(image->directory)+MaxTextExtent+1));
               if (image->directory == (char *) NULL)
-                ThrowReaderException(CorruptImageError,UnableToReadImageData,
+                ThrowMPCReaderException(CorruptImageError,UnableToReadImageData,
                   image);
               p=image->directory+strlen(image->directory);
             }
           c=ReadBlobByte(image);
+          if (c == EOF)
+            break;
           *p++=c;
         } while (c != '\0');
       }
@@ -712,7 +776,7 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
             continue;
           profiles[i].info=MagickAllocateMemory(unsigned char *,profiles[i].length);
           if (profiles[i].info == (unsigned char *) NULL)
-            ThrowReaderException(CorruptImageError,UnableToReadGenericProfile,
+            ThrowMPCReaderException(CorruptImageError,UnableToReadGenericProfile,
               image);
           (void) ReadBlob(image,profiles[i].length,profiles[i].info);
           (void) SetImageProfile(image,profiles[i].name,profiles[i].info,
@@ -730,7 +794,7 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
           Create image colormap.
         */
         if (!AllocateImageColormap(image,image->colors))
-          ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,
+          ThrowMPCReaderException(ResourceLimitError,MemoryAllocationFailed,
             image);
         if (image->colors == 0)
           for (i=0; i < 256; i++)
@@ -754,7 +818,7 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
             packet_size=image->depth > 8 ? 6 : 3;
             colormap=MagickAllocateMemory(unsigned char *,packet_size*image->colors);
             if (colormap == (unsigned char *) NULL)
-              ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,
+              ThrowMPCReaderException(ResourceLimitError,MemoryAllocationFailed,
                 image);
             (void) ReadBlob(image,packet_size*image->colors,colormap);
             p=colormap;
@@ -789,14 +853,14 @@ static Image *ReadMPCImage(const ImageInfo *image_info,ExceptionInfo *exception)
         break;
 
     if (CheckImagePixelLimits(image, exception) != MagickPass)
-      ThrowReaderException(ResourceLimitError,ImagePixelLimitExceeded,image);
+      ThrowMPCReaderException(ResourceLimitError,ImagePixelLimitExceeded,image);
 
     /*
       Attach persistent pixel cache.
     */
     status=PersistCache(image,cache_filename,MagickTrue,&offset,exception);
     if (status == MagickFail)
-      ThrowReaderException(CacheError,UnableToPeristPixelCache,image);
+      ThrowMPCReaderException(CacheError,UnableToPeristPixelCache,image);
     /*
       Proceed to next image.
     */
@@ -868,6 +932,7 @@ ModuleExport void RegisterMPCImage(void)
   entry->encoder=(EncoderHandler) WriteMPCImage;
   entry->magick=(MagickHandler) IsMPC;
   entry->description="Magick Persistent Cache image format";
+  entry->seekable_stream=MagickTrue;
   entry->module="MPC";
   entry->coder_class=UnstableCoderClass;
   (void) RegisterMagickInfo(entry);
@@ -950,16 +1015,19 @@ static MagickPassFail WriteMPCImage(const ImageInfo *image_info,Image *image)
 
   ImageProfileIterator
     profile_iterator;
-  
+
   const char
     *profile_name;
-  
+
   const unsigned char
     *profile_info;
-  
+
   size_t
     profile_length;
 
+  size_t
+    image_list_length;
+
   /*
     Open persistent cache.
   */
@@ -967,6 +1035,7 @@ static MagickPassFail WriteMPCImage(const ImageInfo *image_info,Image *image)
   assert(image_info->signature == MagickSignature);
   assert(image != (Image *) NULL);
   assert(image->signature == MagickSignature);
+  image_list_length=GetImageListLength(image);
   status=OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
   if (status == MagickFail)
     ThrowWriterException(FileOpenError,UnableToOpenFile,image);
@@ -1167,7 +1236,7 @@ static MagickPassFail WriteMPCImage(const ImageInfo *image_info,Image *image)
       }
     /*
       Attached profiles.
-    */    
+    */
     profile_iterator=AllocateImageProfileIterator(image);
     if (profile_iterator)
       {
@@ -1290,7 +1359,7 @@ static MagickPassFail WriteMPCImage(const ImageInfo *image_info,Image *image)
     if (image->next == (Image *) NULL)
       break;
     image=SyncNextImageInList(image);
-    status=MagickMonitorFormatted(scene++,GetImageListLength(image),
+    status=MagickMonitorFormatted(scene++,image_list_length,
                                   &image->exception,SaveImagesText,
                                   image->filename);
     if (status == False)
