@@ -787,29 +787,29 @@ static void LogPALMHeader(const PalmHeader* palm_header)
 
 static const unsigned char *
 GetPalmPaletteGivenBits(const unsigned int bits,
-                        size_t *size)
+                       size_t *entries)
 {
   const unsigned char
     *palette = (const unsigned char *) NULL;
 
-  *size=0;
+  *entries=0;
   switch (bits)
     {
     case 1:
       palette=(const unsigned char *) PalmPalette1;
-      *size=sizeof(PalmPalette1);
+      *entries=sizeof(PalmPalette1)/3U;
       break;
     case 2:
       palette=(const unsigned char *) PalmPalette2;
-      *size=sizeof(PalmPalette2);
+      *entries=sizeof(PalmPalette2)/3U;
       break;
     case 4:
       palette=(const unsigned char *) PalmPalette4;
-      *size=sizeof(PalmPalette4);
+      *entries=sizeof(PalmPalette4)/3U;
       break;
     case 8:
       palette=(const unsigned char *) PalmPalette8;
-      *size=sizeof(PalmPalette8);
+      *entries=sizeof(PalmPalette8)/3U;
       break;
     }
   return palette;
@@ -859,7 +859,7 @@ static Image *ReadPALMImage(const ImageInfo *image_info,
   Image
     *image;
 
-  IndexPacket
+  unsigned int
     index;
 
   long
@@ -882,6 +882,9 @@ static Image *ReadPALMImage(const ImageInfo *image_info,
     *one_row,
     *ptr;
 
+  size_t
+    alloc_size;
+
   unsigned int
     status;
 
@@ -993,141 +996,85 @@ static Image *ReadPALMImage(const ImageInfo *image_info,
   /*
     Initialize image colormap.
   */
-  if ((palm_header.bits_per_pixel < 16) &&
-      !AllocateImageColormap(image,1L << palm_header.bits_per_pixel))
-    ThrowPALMReaderException(ResourceLimitError,MemoryAllocationFailed,image);
-
   if ((palm_header.bits_per_pixel == 1) ||
       (palm_header.bits_per_pixel == 2) ||
       (palm_header.bits_per_pixel == 4) ||
-      ((palm_header.bits_per_pixel == 8) &&
-       !(palm_header.flags & PALM_HAS_COLORMAP_FLAG)))
+      (palm_header.bits_per_pixel == 8))
     {
-      const unsigned char
-        *palette;
-
       size_t
-        size;
+        palette_entries=0;
 
-      palette=GetPalmPaletteGivenBits(palm_header.bits_per_pixel,
-                                      &size);
-      if (palette)
+      const unsigned char
+        *fixed_palette = NULL;
+
+      if (!AllocateImageColormap(image,1UL << palm_header.bits_per_pixel))
+        ThrowPALMReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+
+      if ((palm_header.bits_per_pixel == 8) &&
+          (palm_header.flags & PALM_HAS_COLORMAP_FLAG))
         {
-          if (image->logging)
-            (void) LogMagickEvent(CoderEvent,GetMagickModule(),
-                                  "Default %u bit palette of %"
-                                  MAGICK_SIZE_T_F "u bytes...",
-                                  palm_header.bits_per_pixel,
-                                  (MAGICK_SIZE_T) size);
-          for (index=0; index < image->colors; index++)
+          palette_entries = ReadBlobMSBShort(image);
+
+          if (EOFBlob(image))
+            ThrowPALMReaderException(CorruptImageError,UnexpectedEndOfFile,image);
+          if (palette_entries > image->colors)
+            ThrowPALMReaderException(CorruptImageError,ColormapExceedsColorsLimit,image);
+        }
+
+      if (palette_entries == 0)
+        fixed_palette=GetPalmPaletteGivenBits(palm_header.bits_per_pixel, &palette_entries);
+      if (palette_entries == 0)
+        palette_entries=image->colors;
+
+      if (image->logging)
+        (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                              "%s %u bit palette with %" MAGICK_SIZE_T_F "u colors",
+                              fixed_palette ? "Default" : "Custom",
+                              palm_header.bits_per_pixel,
+                              (MAGICK_SIZE_T) palette_entries);
+
+      if (fixed_palette)
+        {
+          const unsigned char
+            *palette = fixed_palette;
+          for (index=0; index < palette_entries; index++)
             {
               image->colormap[index].red=ScaleCharToQuantum(*palette++);
               image->colormap[index].green=ScaleCharToQuantum(*palette++);
               image->colormap[index].blue=ScaleCharToQuantum(*palette++);
             }
         }
-    }
-
-  else if (palm_header.bits_per_pixel == 8)
-    {
-      i = 0;
-      if (palm_header.flags & PALM_HAS_COLORMAP_FLAG)
+      else if (palm_header.flags & PALM_HAS_COLORMAP_FLAG)
         {
-          unsigned int
-            count;
-
-          count = ReadBlobMSBShort(image);
-
-          if (image->logging)
-            (void) LogMagickEvent(CoderEvent,GetMagickModule(),
-                                  "Custom %u bit palette with %u colors (%"
-                                  MAGICK_SIZE_T_F "u bytes)...",
-                                  palm_header.bits_per_pixel,
-                                  count,
-                                  (MAGICK_SIZE_T) count*3);
-
-          if (count > image->colors)
-            ThrowPALMReaderException(CorruptImageError,ColormapExceedsColorsLimit,image);
-
           /*
             Is palette really written reversed?
           */
-          for (i = 0; i < count; i++)
+          for (i = 0; i < palette_entries; i++)
             {
               /* unsigned int r, g, b; */
               (void) ReadBlobByte(image);
-              index=255 - i;
+              index=(image->colors-1) - i;
               VerifyColormapIndex(image,index);
-#if 0
-              r=ReadBlobByte(image);
-              g=ReadBlobByte(image);
-              b=ReadBlobByte(image);
-              fprintf(stdout,"{ %3u, %3u, %3u },\n", r, g, b);
-#endif
-#if 1
               image->colormap[index].red = ScaleCharToQuantum(ReadBlobByte(image));
               image->colormap[index].green = ScaleCharToQuantum(ReadBlobByte(image));
               image->colormap[index].blue = ScaleCharToQuantum(ReadBlobByte(image));
-#endif
               if (EOFBlob(image))
                 ThrowPALMReaderException(CorruptImageError,UnexpectedEndOfFile,image);
             }
           /*
             Initialize remaining range ???.
           */
-          for (; i < (1U << palm_header.bits_per_pixel); i++)
+          for (; i < image->colors; i++)
             {
-              index=255 - i;
+              index=(image->colors-1) - i;
               image->colormap[index].red = 0;
               image->colormap[index].green = 0;
               image->colormap[index].blue = 0;
             }
         }
-#if 0
-      for (; i < (1U << palm_header.bits_per_pixel); i++)
-        {
-          index=255 - i;
-          VerifyColormapIndex(image,index);
-          image->colormap[index].red = ScaleCharToQuantum(PalmPalette[i][0]);
-          image->colormap[index].green = ScaleCharToQuantum(PalmPalette[i][1]);
-          image->colormap[index].blue = ScaleCharToQuantum(PalmPalette[i][2]);
-        }
-#endif
-#if 0
-      for (i=0; i < (1U << palm_header.bits_per_pixel); i++)
-        {
-          fprintf(stdout,"{ %3u, %3u, %3u },\n",
-                  ScaleQuantumToChar(image->colormap[i].red),
-                  ScaleQuantumToChar(image->colormap[i].green),
-                  ScaleQuantumToChar(image->colormap[i].blue));
-        }
-#endif
     }
 
-image->depth = 8;
-#if 0
-  if (palm_header.bits_per_pixel < 16)
-    {
-      image->storage_class = PseudoClass;
-      image->depth = 8;
-    }
-  else
-    {
-      image->storage_class = DirectClass;
-      image->depth = 8;
-    }
-#endif
-
-#if 0
-  if (image->storage_class == PseudoClass)
-    for (i=0; i < image->colors; i++)
-      fprintf(stderr,"%03u: %3u, %3u, %3u\n",
-              i,
-              ScaleQuantumToChar(image->colormap[i].red),
-              ScaleQuantumToChar(image->colormap[i].green),
-              ScaleQuantumToChar(image->colormap[i].blue));
-#endif
-
+  image->depth = 8;
   image->compression = NoCompression;
   if (palm_header.flags & PALM_IS_COMPRESSED_FLAG)
     {
@@ -1149,16 +1096,18 @@ image->depth = 8;
   if (CheckImagePixelLimits(image, exception) != MagickPass)
     ThrowPALMReaderException(ResourceLimitError,ImagePixelLimitExceeded,image);
 
-  one_row = MagickAllocateMemory(unsigned char *,Max(palm_header.bytes_per_row,
-                                                     MagickArraySize(2,image->columns)));
+  alloc_size = Max(palm_header.bytes_per_row,MagickArraySize(2,image->columns));
+  one_row = MagickAllocateMemory(unsigned char *,alloc_size);
   if (one_row == (unsigned char *) NULL)
     ThrowPALMReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+  (void) memset(one_row,0,alloc_size);
   if (palm_header.compression_type == PALM_COMPRESSION_SCANLINE)
     {
-      lastrow = MagickAllocateMemory(unsigned char *,Max(palm_header.bytes_per_row,
-                                                         MagickArraySize(2,image->columns)));
+      alloc_size = Max(palm_header.bytes_per_row,MagickArraySize(2,image->columns));
+      lastrow = MagickAllocateMemory(unsigned char *,alloc_size);
       if (lastrow == (unsigned char *) NULL)
         ThrowPALMReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+      (void) memset(lastrow,0,alloc_size);
     }
 
   mask = (1l << palm_header.bits_per_pixel) - 1;
@@ -1229,7 +1178,7 @@ image->depth = 8;
       indexes=AccessMutableIndexes(image);
       if (palm_header.bits_per_pixel == 16)
         {
-          if (image->columns > 2*palm_header.bytes_per_row)
+          if (image->columns > 2U*palm_header.bytes_per_row)
             ThrowPALMReaderException(CorruptImageError,CorruptImage,image);
           for (x=0; x < (long) image->columns; x++)
             {
